-- | `GenericParser.DomainParser.Common` regroups all functions and types that are common between the different domain parsers.
module GenericParser.DomainParser.Common where

import Control.Alt ((<|>))
import Data.Maybe (Maybe(..))
import Data.String as S

import GenericParser.Parser (Parser(..)
                            , success, failureError
                            , char, many1)
import GenericParser.SomeParsers (alphanum)

type Size = Int
-- | `DomainError` expresses all possible errors that can occur while parsing a domain.
-- | When an error occurs, the position is given by the Parser along the related `DomainError`.
data DomainError
  = LabelTooLarge Size
  | DomainTooLarge Size
  | InvalidCharacter
  | EOFExpected

-- | From RFC 1035: <let-dig> ::= <letter> | <digit>
let_dig :: forall e. Parser e Char
let_dig = alphanum

-- | From RFC 1035: <let-dig-hyp> ::= <let-dig> | "-"
-- | Either a Letter, Digital or an Hyphenation character.
let_dig_hyp :: forall e. Parser e Char
let_dig_hyp = let_dig <|> char '-'

-- | From RFC 1035: <ldh-str> ::= <let-dig-hyp> | <let-dig-hyp> <ldh-str>
ldh_str :: forall e. Parser e (Array Char)
ldh_str = many1 let_dig_hyp

-- | WARNING: Verify the actual maximum length for a label.
-- | Current maximum accepted length for a label is 63.
max_label_length :: Int
max_label_length = 63

-- | WARNING: Verify the actual maximum length for a domain.
-- | Current maximum accepted length for a domain is 255.
max_domain_length :: Int
max_domain_length = 255

-- | Test for the end-of-file (no more input).
-- | If not EOF the parser fails (Nothing), otherwise it provides an empty string.
eof :: Parser DomainError String
eof = Parser \input -> case S.length input.string of
  0 -> success input ""
  _ -> failureError input.position (Just EOFExpected)
