-- | `RFC5234`, Augmented BNF notation, explains the syntax specification found in many RFCs.
-- | This module implements core rules found in appendix B.1.
module GenericParser.RFC5234 where

import Prelude (between, pure, (<$>), (<<<), (||), (==), bind, ($))

import Control.Alt ((<|>))
import Data.Array as A
import Data.Char as C
import Data.String.CodeUnits as CU

import GenericParser.BaseFunctions (isAlpha, isDigit, isHexaDecimal)

import GenericParser.Parser (Parser, char, sat, item)

-- | ALPHA: any letter, upper or lower case.
-- |
-- | ALPHA          =  %x41-5A / %x61-7A   ; A-Z / a-z
alpha :: forall e. Parser e Char
alpha = sat isAlpha

-- | BIT: either character '0' or '1'.
-- |
-- | BIT            =  "0" / "1"
bit :: forall e. Parser e Char
bit = char '0' <|> char '1'

-- | CHAR (renamed `asciichar` to fix naming conflict with `GenericParser.char`):
-- | any 7-bit US-ASCII character, excluding NUL.
-- |
-- | CHAR           =  %x01-7F
asciichar :: forall e. Parser e Char
asciichar = sat (between 1 127 <<< C.toCharCode)

-- | CR: carriage return.
-- |
-- | CR             =  %x0D
cr :: forall e. Parser e Char
cr = char '\r'

-- | CRLF: Internet standard newline.
-- |
-- | CRLF           =  CR LF
crlf :: forall e. Parser e String
crlf = do _ <- char '\r'
          _ <- char '\n'
          pure "\r\n"

-- | CTL: control characters.
-- |
-- | CTL            =  %x00-1F / %x7F
ctl :: forall e. Parser e Char
ctl = sat cond
  where cond x = (between 0 31 $ C.toCharCode x)
                 || C.toCharCode x == 127

-- | DIGIT: any digit character (from 0 to 9).
-- |
-- | DIGIT          =  %x30-39
digit :: forall e. Parser e Char
digit = sat isDigit

-- | DQUOTE: double quote (").
-- |
-- | DQUOTE         =  %x22
dquote :: forall e. Parser e Char
dquote = sat (\x -> C.toCharCode x == 34)

-- | HEXDIG: hexadecimal.
-- |
-- | HEXDIG         =  DIGIT / "A" / "B" / "C" / "D" / "E" / "F"
hexdig :: forall e. Parser e Char
hexdig = sat isHexaDecimal

-- | HTAB: horizontal tab.
-- | HTAB           =  %x09
htab :: forall e. Parser e Char
htab = char '\t'

-- | LF: linefeed.
-- |
-- | LF             =  %x0A
lf :: forall e. Parser e Char
lf = char '\n'

-- | LWSP: Use of this linear-white-space rule permits lines containing only white
-- | space that are no longer legal in mail headers and have caused interoperability
-- | problems in other contexts.
-- | Do not use when defining mail headers and use with caution in other contexts.
-- |
-- | LWSP           =  *(WSP / CRLF WSP)
lwsp :: forall e. Parser e String
lwsp = CU.fromCharArray <$> A.many (wsp <|> (do _ <- crlf
                                                wsp))

-- | OCTET: any data on a single byte.
-- |
-- | OCTET          =  %x00-FF
octet :: forall e. Parser e Char
octet = item

-- | SP: space.
-- |
-- | SP             =  %x20
sp :: forall e. Parser e Char
sp = char ' '

-- | VCHAR          =  %x21-7E
-- |
-- | Visible printing characters.
vchar :: forall e. Parser e Char
vchar = sat (between 33 126 <<< C.toCharCode)

-- | WSP: white space.
-- |
-- | WSP            =  SP / HTAB
wsp :: forall e. Parser e Char
wsp = sp <|> htab
